{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE RankNTypes #-}
module Data.Machine.Seekable
 ( Seekable(..)
 , fullRead
 , seeks
 , seekByteStringHandle
 , seekByteString
 , seeking
 ) where

import           Control.Monad
import           Control.Monad.IO.Class
import qualified Data.ByteString as BS
import           Data.ByteString (ByteString)
import           Data.Machine
import           Data.Word
import           System.IO (Handle, hSeek, SeekMode(..))


data Seekable i a b where
  SeekPos :: i -> Seekable i a ()
  Read    :: Seekable i a a

-- | Streams the remaining data in the seekable.
fullRead :: Monad m => MachineT m (Seekable i a) a
fullRead = construct . forever $ awaits Read >>= yield

-- | Compose two seeking machines where you can seek somewhere from a lower machine
--   this completely stops breaks the upstream machine if there's any internal partial state
seeks :: Monad m => (i2 -> i1) -> MachineT m (Seekable i1 a) b -> MachineT m (Seekable i2 b) c -> MachineT m (Seekable i1 a) c
seeks iF up down =
  stepMachine down $ \case
    Stop -> encased Stop
    Yield x down' -> encased $ Yield x $ seeks iF up down'
    downAwait@(Await down' Read fDown) ->
      stepMachine up $ \case
        Stop -> encased Stop
        Yield y up' -> seeks iF up' $ down' y
        Await up' Read fUp ->
          encased $ Await (\i -> seeks iF (up' i) (encased downAwait)) Read $ seeks iF fUp fDown
        -- Perform the seek and then continue with upstream with downstream still at the await of a read
        Await up' (SeekPos p) fUp ->
          encased $ Await (\i -> seeks iF (up' i) $ encased downAwait) (SeekPos p) $ seeks iF fUp fDown
    Await down' (SeekPos pDown) fDown ->
      encased $ Await (\i -> seeks iF up (down' i)) (SeekPos $ iF pDown) $ seeks iF up fDown


-- | Treats a handle readable as a ByteString as a seekable machine.
--   The handle must actually be seekable.
seekByteStringHandle :: MonadIO m => Word -> Handle -> MachineT m (Seekable Word64 ByteString) a -> SourceT m a
seekByteStringHandle chunkSize' h = seeking seekTo initial
  where
    chunkSize = fromIntegral chunkSize'
    seekTo i = MachineT (liftIO (hSeek h AbsoluteSeek (fromIntegral i)) >> runMachineT initial)
    initial :: MonadIO m => SourceT m ByteString
    initial = MachineT $ do
      o <- liftIO (BS.hGet h chunkSize)
      if BS.null o
        then pure Stop
        else pure $ Yield o initial

-- | Represents as descrete bytestring as a seekable source of bytestrings.
--
--   Usefull both for testing and mmapped data.
seekByteString :: Monad m => ByteString -> MachineT m (Seekable Word64 ByteString) a -> SourceT m a
seekByteString bs = seeking seekTo initial
  where
    seekTo i = construct $ yield $ BS.drop (fromIntegral i) bs
    initial = construct $ yield bs

-- | A constructor of Seekable Machines.
--   It takes a function that produces a machine reading from a given point, the initial machine,
--   and the seekable machine to run.
--   From that it produces a machine that does the seeking internally.
seeking :: Monad m => (i -> MachineT m k a) -> MachineT m k a -> MachineT m (Seekable i a) o -> MachineT m k o
seeking mkAtPnt initUp down = stepMachine down $ \case
  Stop -> stopped
  Yield o down' -> encased (Yield o (seeking mkAtPnt initUp down'))
  Await down' (SeekPos pnt) _ -> seeking mkAtPnt (mkAtPnt pnt) (down' ())
  stepD@(Await down' (Read) ffD) -> stepMachine initUp $ \case
                               Stop -> seeking mkAtPnt stopped ffD
                               Yield o up' -> seeking mkAtPnt up' (down' o)
                               Await up' req ffU -> encased (Await (\a -> seeking mkAtPnt (up' a) (encased stepD))
                                                             req
                                                             (seeking mkAtPnt ffU (encased stepD))) 
